submission 505727
gau.nernst · python · License unknown
Kernel source · 1262 lines ↓holds 1 record
Use it
Vendorable · source mirrored · license unknownView source →
No package. Vendor the mirrored source: 1262 lines, June 9 Researcher Reciprocity License v1.0.
submission_v6c.py
curl "https://kernelindex.com/api/v1/implementations/kernelbot-nvfp4-group-gemm-505727?include=source"interfacepython
Compatibility
measured onNVIDIA B200
declared hardwareNVIDIA B200
architecturessm_100
dtypesfp8_e4m3, nvfp4
Benchmark evidence
1 measurement across 1 GPU, fastest first.
Operation / workload
Hardware
Latency
Rank
Observed
Reported · How evidence levels are derived →
Source and license
sourceavailable
revision digestsha256:2f511095e0b6d3e31c2ac40c6ef28ffb23e6b68969a80c06e307dde8c68f2719
license declaredunknown
license concludedunknown
authorsgau.nernst
imported2026-08-15
Techniques
Extracted from the mirrored source by pattern, never inferred. Each row cites its line.
cluster
__cluster_dims__(CTA_GROUP, 1, 1)fused-epilogue
const int epilogue_mbar_addr = mainloop_mbar_addr + 3 * 8;mbarrier
void mbarrier_init(int mbar_addr, int count) {num-warps = 6
constexpr int NUM_WARPS = 6;shared-memory
extern __shared__ __align__(1024) char smem_ptr[];tcgen05
asm volatile("tcgen05.cp.cta_group::%2.32x128b.warpx4 [%0], %1;" :: "r"(taddr), "l"(s_desc), "n"(CTA_GROUP));tile-k = 256
constexpr int BLOCK_K = 256;tile-m = 128
issue_SFA = BLOCK_M == 128 || (bid_m * BLOCK_M + cta_rank * 128 < M); // when BLOCK_M=128, we must always issue SFA TMAtile-n = 128
constexpr int BLOCK_N = 128;tma
asm volatile("cp.async.bulk.prefetch.L2.global.L2::cache_hint [%0], %1, %2;"Kernel source
submission_v6c.py1262 lines
#!POPCORN leaderboard nvfp4_group_gemm
#!POPCORN gpu B200
import torch
from task import input_t, output_t
from torch.utils.cpp_extension import load_inline
CUDA_SRC = r"""
#include <cudaTypedefs.h>
#include <cuda_fp16.h>
#include <torch/library.h>
#include <ATen/core/Tensor.h>
constexpr int WARP_SIZE = 32;
constexpr int MMA_K = 64; // 32 bytes
__device__ __host__
constexpr int cdiv(int a, int b) { return (a + b - 1) / b; }
// https://github.com/NVIDIA/cutlass/blob/v4.3.2/include/cute/arch/copy_sm90_desc.hpp#L193-L197
constexpr uint64_t EVICT_NORMAL = 0x1000000000000000;
constexpr uint64_t EVICT_FIRST = 0x12F0000000000000;
constexpr uint64_t EVICT_LAST = 0x14F0000000000000;
__device__ inline
constexpr uint64_t desc_encode(uint64_t x) { return (x & 0x3'FFFFULL) >> 4ULL; };
// https://github.com/NVIDIA/cutlass/blob/v4.2.1/include/cute/arch/cluster_sm90.hpp#L180
__device__
uint32_t elect_sync() {
uint32_t pred = 0;
asm volatile(
"{\n\t"
".reg .pred %%px;\n\t"
"elect.sync _|%%px, %1;\n\t"
"@%%px mov.s32 %0, 1;\n\t"
"}"
: "+r"(pred)
: "r"(0xFFFFFFFF)
);
return pred;
}
// named barrier
template <int bar>
__device__ inline
void bar_sync(int count) {
asm volatile("bar.sync %0, %1;" :: "n"(bar), "r"(count) : "memory");
}
__device__ inline
void mbarrier_init(int mbar_addr, int count) {
asm volatile("mbarrier.init.shared::cta.b64 [%0], %1;" :: "r"(mbar_addr), "r"(count));
}
__device__ inline
void mbarrier_arrive(int mbar_addr) {
asm volatile("mbarrier.arrive.release.cta.shared::cluster.b64 _, [%0];" :: "r"(mbar_addr) : "memory");
}
// NOTE: using .shared::cluster
__device__ inline
void mbarrier_arrive_expect_tx(int mbar_addr, int size) {
asm volatile("mbarrier.arrive.expect_tx.release.cta.shared::cluster.b64 _, [%0], %1;" :: "r"(mbar_addr), "r"(size) : "memory");
}
// https://github.com/NVIDIA/cutlass/blob/v4.2.1/include/cutlass/arch/barrier.h#L408
__device__
void mbarrier_wait(int mbar_addr, int phase) {
uint32_t ticks = 0x989680; // this is optional
asm volatile(
"{\n\t"
".reg .pred P1;\n\t"
"LAB_WAIT:\n\t"
"mbarrier.try_wait.parity.acquire.cta.shared::cta.b64 P1, [%0], %1, %2;\n\t"
"@!P1 bra.uni LAB_WAIT;\n\t"
"}"
:: "r"(mbar_addr), "r"(phase), "r"(ticks)
);
}
__device__ inline
void prefetch_tensormap(const void *tmap_ptr) {
asm volatile("prefetch.tensormap [%0];" :: "l"(tmap_ptr));
}
__device__ inline
void tma_prefetch(const void *src, int size, uint64_t cache_policy) {
asm volatile("cp.async.bulk.prefetch.L2.global.L2::cache_hint [%0], %1, %2;"
:: "l"(src), "r"(size), "l"(cache_policy) : "memory");
}
__device__ inline
void tma_1d_prefetch(const void *tmap_ptr, int x, uint64_t cache_policy) {
asm volatile("cp.async.bulk.prefetch.tensor.1d.L2.global.L2::cache_hint [%0, {%1}], %2;"
:: "l"(tmap_ptr), "r"(x), "l"(cache_policy) : "memory");
}
__device__ inline
void tma_2d_prefetch(const void *tmap_ptr, int x, int y, uint64_t cache_policy) {
asm volatile("cp.async.bulk.prefetch.tensor.2d.L2.global.L2::cache_hint [%0, {%1, %2}], %3;"
:: "l"(tmap_ptr), "r"(x), "r"(y), "l"(cache_policy) : "memory");
}
__device__ inline
void tma_3d_prefetch(const void *tmap_ptr, int x, int y, int z, uint64_t cache_policy) {
asm volatile("cp.async.bulk.prefetch.tensor.3d.L2.global.L2::cache_hint [%0, {%1, %2, %3}], %4;"
:: "l"(tmap_ptr), "r"(x), "r"(y), "r"(z), "l"(cache_policy) : "memory");
}
__device__ inline
void tma_g2s(int dst, const void *src, int size, int mbar_addr, uint64_t cache_policy) {
asm volatile("cp.async.bulk.shared::cluster.global.mbarrier::complete_tx::bytes.L2::cache_hint [%0], [%1], %2, [%3], %4;"
:: "r"(dst), "l"(src), "r"(size), "r"(mbar_addr), "l"(cache_policy));
}
template <int CTA_GROUP = 1>
__device__ inline
void tma_1d_g2s(int dst, const void *tmap_ptr, int x, int mbar_addr, uint64_t cache_policy) {
asm volatile("cp.async.bulk.tensor.1d.shared::cluster.global.mbarrier::complete_tx::bytes.cta_group::%5.L2::cache_hint "
"[%0], [%1, {%2}], [%3], %4;"
:: "r"(dst), "l"(tmap_ptr), "r"(x), "r"(mbar_addr), "l"(cache_policy), "n"(CTA_GROUP)
: "memory");
}
template <int CTA_GROUP = 1>
__device__ inline
void tma_1d_g2s_mcast(int dst, const void *tmap_ptr, int x, int mbar_addr, int16_t cta_mask, uint64_t cache_policy) {
asm volatile("cp.async.bulk.tensor.1d.shared::cluster.global.mbarrier::complete_tx::bytes.multicast::cluster.cta_group::%6.L2::cache_hint "
"[%0], [%1, {%2}], [%3], %4, %5;"
:: "r"(dst), "l"(tmap_ptr), "r"(x), "r"(mbar_addr), "h"(cta_mask), "l"(cache_policy), "n"(CTA_GROUP)
: "memory");
}
template <int CTA_GROUP = 1>
__device__ inline
void tma_2d_g2s(int dst, const void *tmap_ptr, int x, int y, int mbar_addr, uint64_t cache_policy) {
asm volatile("cp.async.bulk.tensor.2d.shared::cluster.global.mbarrier::complete_tx::bytes.cta_group::%6.L2::cache_hint "
"[%0], [%1, {%2, %3}], [%4], %5;"
:: "r"(dst), "l"(tmap_ptr), "r"(x), "r"(y), "r"(mbar_addr), "l"(cache_policy), "n"(CTA_GROUP)
: "memory");
}
template <int CTA_GROUP = 1>
__device__ inline
void tma_2d_g2s_mcast(int dst, const void *tmap_ptr, int x, int y, int mbar_addr, int16_t cta_mask, uint64_t cache_policy) {
asm volatile("cp.async.bulk.tensor.2d.shared::cluster.global.mbarrier::complete_tx::bytes.multicast::cluster.cta_group::%7.L2::cache_hint "
"[%0], [%1, {%2, %3}], [%4], %5, %6;"
:: "r"(dst), "l"(tmap_ptr), "r"(x), "r"(y), "r"(mbar_addr), "h"(cta_mask), "l"(cache_policy), "n"(CTA_GROUP)
: "memory");
}
template <int CTA_GROUP = 1>
__device__ inline
void tma_3d_g2s(int dst, const void *tmap_ptr, int x, int y, int z, int mbar_addr, uint64_t cache_policy) {
asm volatile("cp.async.bulk.tensor.3d.shared::cluster.global.mbarrier::complete_tx::bytes.cta_group::%7.L2::cache_hint "
"[%0], [%1, {%2, %3, %4}], [%5], %6;"
:: "r"(dst), "l"(tmap_ptr), "r"(x), "r"(y), "r"(z), "r"(mbar_addr), "l"(cache_policy), "n"(CTA_GROUP)
: "memory");
}
template <int CTA_GROUP = 1>
__device__ inline
void tma_3d_g2s_mcast(int dst, const void *tmap_ptr, int x, int y, int z, int mbar_addr, int16_t cta_mask, uint64_t cache_policy) {
asm volatile("cp.async.bulk.tensor.3d.shared::cluster.global.mbarrier::complete_tx::bytes.multicast::cluster.cta_group::%8.L2::cache_hint "
"[%0], [%1, {%2, %3, %4}], [%5], %6, %7;"
:: "r"(dst), "l"(tmap_ptr), "r"(x), "r"(y), "r"(z), "r"(mbar_addr), "h"(cta_mask), "l"(cache_policy), "n"(CTA_GROUP)
: "memory");
}
template <int CTA_GROUP = 1>
__device__ inline
void tcgen05_cp_nvfp4(int taddr, uint64_t s_desc) {
// .32x128b corresponds to (32, 16) 8-bit scale -> 1 MMA for nvfp4.
// .warpx4 duplicates data across 32-lane groups.
asm volatile("tcgen05.cp.cta_group::%2.32x128b.warpx4 [%0], %1;" :: "r"(taddr), "l"(s_desc), "n"(CTA_GROUP));
}
template <int CTA_GROUP = 1>
__device__ inline
void tcgen05_alloc(int addr, int size) {
asm volatile("tcgen05.alloc.cta_group::%2.sync.aligned.shared::cta.b32 [%0], %1;" :: "r"(addr), "r"(size), "n"(CTA_GROUP));
}
template <int CTA_GROUP = 1>
__device__ inline
void tcgen05_dealloc(int addr, int size) {
asm volatile("tcgen05.dealloc.cta_group::%2.sync.aligned.b32 %0, %1;" :: "r"(0), "r"(size), "n"(CTA_GROUP));
}
template <int CTA_GROUP = 1>
__device__ inline
void tcgen05_commit(int mbar_addr) {
asm volatile("tcgen05.commit.cta_group::%1.mbarrier::arrive::one.shared::cluster.b64 [%0];"
:: "r"(mbar_addr), "n"(CTA_GROUP) : "memory");
}
template <int CTA_GROUP = 1>
__device__ inline
void tcgen05_commit_mcast(int mbar_addr, uint16_t cta_mask) {
asm volatile("tcgen05.commit.cta_group::%2.mbarrier::arrive::one.shared::cluster.multicast::cluster.b64 [%0], %1;"
:: "r"(mbar_addr), "h"(cta_mask), "n"(CTA_GROUP) : "memory");
}
struct COLLECTOR_USAGE {
static constexpr char NONE[] = "";
static constexpr char A_FILL[] = ".collector::a::fill";
static constexpr char A_USE[] = ".collector::a::use";
static constexpr char A_LASTUSE[] = ".collector::a::lastuse";
static constexpr char A_DISCARD[] = ".collector::a::discard";
};
template <int CTA_GROUP = 1, const char *collector_usage = COLLECTOR_USAGE::NONE>
__device__ inline
void tcgen05_mma_nvfp4(
int d_tmem,
uint64_t a_desc,
uint64_t b_desc,
uint32_t i_desc,
int scale_A_tmem,
int scale_B_tmem,
int enable_input_d
) {
asm volatile(
"{\n\t"
".reg .pred p;\n\t" // predicate register enable-input-d
"setp.ne.b32 p, %6, 0;\n\t"
"tcgen05.mma.cta_group::%7.kind::mxf4nvf4.block_scale.block16%8 [%0], %1, %2, %3, [%4], [%5], p;\n\t"
"}"
:: "r"(d_tmem), "l"(a_desc), "l"(b_desc), "r"(i_desc),
"r"(scale_A_tmem), "r"(scale_B_tmem), "r"(enable_input_d),
"n"(CTA_GROUP), "C"(collector_usage)
);
}
// see https://docs.nvidia.com/cuda/inline-ptx-assembly/index.html
struct SHAPE {
static constexpr char _32x32b[] = ".32x32b"; // 32x1 tile for each warp
static constexpr char _16x128b[] = ".16x128b"; // 16x4 tile
static constexpr char _16x256b[] = ".16x256b"; // 16x8 tile
};
template <int NUM_REGS, const char *SHAPE, int NUM>
__device__ inline
void tcgen05_ld(float *tmp, int row, int col) {
int addr = (row << 16) | col;
if constexpr (NUM_REGS == 1)
asm volatile("tcgen05.ld.sync.aligned%2.x%3.b32 {%0}, [%1];"
: "=f"(tmp[0]) : "r"(addr), "C"(SHAPE), "n"(NUM));
if constexpr (NUM_REGS == 2)
asm volatile("tcgen05.ld.sync.aligned%3.x%4.b32 {%0, %1}, [%2];"
: "=f"(tmp[0]), "=f"(tmp[1]) : "r"(addr), "C"(SHAPE), "n"(NUM));
if constexpr (NUM_REGS == 4)
asm volatile("tcgen05.ld.sync.aligned%5.x%6.b32 "
"{%0, %1, %2, %3}, [%4];"
: "=f"(tmp[0]), "=f"(tmp[1]), "=f"(tmp[2]), "=f"(tmp[3])
: "r"(addr), "C"(SHAPE), "n"(NUM));
if constexpr (NUM_REGS == 8)
asm volatile("tcgen05.ld.sync.aligned%9.x%10.b32 "
"{ %0, %1, %2, %3, %4, %5, %6, %7}, [%8];"
: "=f"(tmp[0]), "=f"(tmp[1]), "=f"(tmp[2]), "=f"(tmp[3]), "=f"(tmp[4]), "=f"(tmp[5]), "=f"(tmp[6]), "=f"(tmp[7])
: "r"(addr), "C"(SHAPE), "n"(NUM));
if constexpr (NUM_REGS == 16)
asm volatile("tcgen05.ld.sync.aligned%17.x%18.b32 "
"{ %0, %1, %2, %3, %4, %5, %6, %7, "
" %8, %9, %10, %11, %12, %13, %14, %15}, [%16];"
: "=f"(tmp[ 0]), "=f"(tmp[ 1]), "=f"(tmp[ 2]), "=f"(tmp[ 3]), "=f"(tmp[ 4]), "=f"(tmp[ 5]), "=f"(tmp[ 6]), "=f"(tmp[ 7]),
"=f"(tmp[ 8]), "=f"(tmp[ 9]), "=f"(tmp[10]), "=f"(tmp[11]), "=f"(tmp[12]), "=f"(tmp[13]), "=f"(tmp[14]), "=f"(tmp[15])
: "r"(addr), "C"(SHAPE), "n"(NUM));
if constexpr (NUM_REGS == 32)
asm volatile("tcgen05.ld.sync.aligned%33.x%34.b32 "
"{ %0, %1, %2, %3, %4, %5, %6, %7, "
" %8, %9, %10, %11, %12, %13, %14, %15, "
" %16, %17, %18, %19, %20, %21, %22, %23, "
" %24, %25, %26, %27, %28, %29, %30, %31}, [%32];"
: "=f"(tmp[ 0]), "=f"(tmp[ 1]), "=f"(tmp[ 2]), "=f"(tmp[ 3]), "=f"(tmp[ 4]), "=f"(tmp[ 5]), "=f"(tmp[ 6]), "=f"(tmp[ 7]),
"=f"(tmp[ 8]), "=f"(tmp[ 9]), "=f"(tmp[10]), "=f"(tmp[11]), "=f"(tmp[12]), "=f"(tmp[13]), "=f"(tmp[14]), "=f"(tmp[15]),
"=f"(tmp[16]), "=f"(tmp[17]), "=f"(tmp[18]), "=f"(tmp[19]), "=f"(tmp[20]), "=f"(tmp[21]), "=f"(tmp[22]), "=f"(tmp[23]),
"=f"(tmp[24]), "=f"(tmp[25]), "=f"(tmp[26]), "=f"(tmp[27]), "=f"(tmp[28]), "=f"(tmp[29]), "=f"(tmp[30]), "=f"(tmp[31])
: "r"(addr), "C"(SHAPE), "n"(NUM));
if constexpr (NUM_REGS == 64)
asm volatile("tcgen05.ld.sync.aligned%65.x%66.b32 "
"{ %0, %1, %2, %3, %4, %5, %6, %7, "
" %8, %9, %10, %11, %12, %13, %14, %15, "
" %16, %17, %18, %19, %20, %21, %22, %23, "
" %24, %25, %26, %27, %28, %29, %30, %31, "
" %32, %33, %34, %35, %36, %37, %38, %39, "
" %40, %41, %42, %43, %44, %45, %46, %47, "
" %48, %49, %50, %51, %52, %53, %54, %55, "
" %56, %57, %58, %59, %60, %61, %62, %63}, [%64];"
: "=f"(tmp[ 0]), "=f"(tmp[ 1]), "=f"(tmp[ 2]), "=f"(tmp[ 3]), "=f"(tmp[ 4]), "=f"(tmp[ 5]), "=f"(tmp[ 6]), "=f"(tmp[ 7]),
"=f"(tmp[ 8]), "=f"(tmp[ 9]), "=f"(tmp[10]), "=f"(tmp[11]), "=f"(tmp[12]), "=f"(tmp[13]), "=f"(tmp[14]), "=f"(tmp[15]),
"=f"(tmp[16]), "=f"(tmp[17]), "=f"(tmp[18]), "=f"(tmp[19]), "=f"(tmp[20]), "=f"(tmp[21]), "=f"(tmp[22]), "=f"(tmp[23]),
"=f"(tmp[24]), "=f"(tmp[25]), "=f"(tmp[26]), "=f"(tmp[27]), "=f"(tmp[28]), "=f"(tmp[29]), "=f"(tmp[30]), "=f"(tmp[31]),
"=f"(tmp[32]), "=f"(tmp[33]), "=f"(tmp[34]), "=f"(tmp[35]), "=f"(tmp[36]), "=f"(tmp[37]), "=f"(tmp[38]), "=f"(tmp[39]),
"=f"(tmp[40]), "=f"(tmp[41]), "=f"(tmp[42]), "=f"(tmp[43]), "=f"(tmp[44]), "=f"(tmp[45]), "=f"(tmp[46]), "=f"(tmp[47]),
"=f"(tmp[48]), "=f"(tmp[49]), "=f"(tmp[50]), "=f"(tmp[51]), "=f"(tmp[52]), "=f"(tmp[53]), "=f"(tmp[54]), "=f"(tmp[55]),
"=f"(tmp[56]), "=f"(tmp[57]), "=f"(tmp[58]), "=f"(tmp[59]), "=f"(tmp[60]), "=f"(tmp[61]), "=f"(tmp[62]), "=f"(tmp[63])
: "r"(addr), "C"(SHAPE), "n"(NUM));
if constexpr (NUM_REGS == 128)
asm volatile("tcgen05.ld.sync.aligned%129.x%130.b32 "
"{ %0, %1, %2, %3, %4, %5, %6, %7, "
" %8, %9, %10, %11, %12, %13, %14, %15, "
" %16, %17, %18, %19, %20, %21, %22, %23, "
" %24, %25, %26, %27, %28, %29, %30, %31, "
" %32, %33, %34, %35, %36, %37, %38, %39, "
" %40, %41, %42, %43, %44, %45, %46, %47, "
" %48, %49, %50, %51, %52, %53, %54, %55, "
" %56, %57, %58, %59, %60, %61, %62, %63, "
" %64, %65, %66, %67, %68, %69, %70, %71, "
" %72, %73, %74, %75, %76, %77, %78, %79, "
" %80, %81, %82, %83, %84, %85, %86, %87, "
" %88, %89, %90, %91, %92, %93, %94, %95, "
" %96, %97, %98, %99,%100,%101,%102,%103, "
"%104,%105,%106,%107,%108,%109,%110,%111, "
"%112,%113,%114,%115,%116,%117,%118,%119, "
"%120,%121,%122,%123,%124,%125,%126,%127}, [%128];"
: "=f"(tmp[ 0]), "=f"(tmp[ 1]), "=f"(tmp[ 2]), "=f"(tmp[ 3]), "=f"(tmp[ 4]), "=f"(tmp[ 5]), "=f"(tmp[ 6]), "=f"(tmp[ 7]),
"=f"(tmp[ 8]), "=f"(tmp[ 9]), "=f"(tmp[10]), "=f"(tmp[11]), "=f"(tmp[12]), "=f"(tmp[13]), "=f"(tmp[14]), "=f"(tmp[15]),
"=f"(tmp[16]), "=f"(tmp[17]), "=f"(tmp[18]), "=f"(tmp[19]), "=f"(tmp[20]), "=f"(tmp[21]), "=f"(tmp[22]), "=f"(tmp[23]),
"=f"(tmp[24]), "=f"(tmp[25]), "=f"(tmp[26]), "=f"(tmp[27]), "=f"(tmp[28]), "=f"(tmp[29]), "=f"(tmp[30]), "=f"(tmp[31]),
"=f"(tmp[32]), "=f"(tmp[33]), "=f"(tmp[34]), "=f"(tmp[35]), "=f"(tmp[36]), "=f"(tmp[37]), "=f"(tmp[38]), "=f"(tmp[39]),
"=f"(tmp[40]), "=f"(tmp[41]), "=f"(tmp[42]), "=f"(tmp[43]), "=f"(tmp[44]), "=f"(tmp[45]), "=f"(tmp[46]), "=f"(tmp[47]),
"=f"(tmp[48]), "=f"(tmp[49]), "=f"(tmp[50]), "=f"(tmp[51]), "=f"(tmp[52]), "=f"(tmp[53]), "=f"(tmp[54]), "=f"(tmp[55]),
"=f"(tmp[56]), "=f"(tmp[57]), "=f"(tmp[58]), "=f"(tmp[59]), "=f"(tmp[60]), "=f"(tmp[61]), "=f"(tmp[62]), "=f"(tmp[63]),
"=f"(tmp[64]), "=f"(tmp[65]), "=f"(tmp[66]), "=f"(tmp[67]), "=f"(tmp[68]), "=f"(tmp[69]), "=f"(tmp[70]), "=f"(tmp[71]),
"=f"(tmp[72]), "=f"(tmp[73]), "=f"(tmp[74]), "=f"(tmp[75]), "=f"(tmp[76]), "=f"(tmp[77]), "=f"(tmp[78]), "=f"(tmp[79]),
"=f"(tmp[80]), "=f"(tmp[81]), "=f"(tmp[82]), "=f"(tmp[83]), "=f"(tmp[84]), "=f"(tmp[85]), "=f"(tmp[86]), "=f"(tmp[87]),
"=f"(tmp[88]), "=f"(tmp[89]), "=f"(tmp[90]), "=f"(tmp[91]), "=f"(tmp[92]), "=f"(tmp[93]), "=f"(tmp[94]), "=f"(tmp[95]),
"=f"(tmp[96]), "=f"(tmp[97]), "=f"(tmp[98]), "=f"(tmp[99]), "=f"(tmp[100]),"=f"(tmp[101]),"=f"(tmp[102]),"=f"(tmp[103]),
"=f"(tmp[104]),"=f"(tmp[105]),"=f"(tmp[106]),"=f"(tmp[107]),"=f"(tmp[108]),"=f"(tmp[109]),"=f"(tmp[110]),"=f"(tmp[111]),
"=f"(tmp[112]),"=f"(tmp[113]),"=f"(tmp[114]),"=f"(tmp[115]),"=f"(tmp[116]),"=f"(tmp[117]),"=f"(tmp[118]),"=f"(tmp[119]),
"=f"(tmp[120]),"=f"(tmp[121]),"=f"(tmp[122]),"=f"(tmp[123]),"=f"(tmp[124]),"=f"(tmp[125]),"=f"(tmp[126]),"=f"(tmp[127])
: "r"(addr), "C"(SHAPE), "n"(NUM));
}
template <int num>
__device__ inline void
tcgen05_ld_32x32b(float *tmp, int row, int col) {
// each 32x32b tile uses 1 register per thread
tcgen05_ld<num, SHAPE::_32x32b, num>(tmp, row, col);
}
template <int num>
__device__ inline
void tcgen05_ld_16x128b(float *tmp, int row, int col) {
// each 16x128b tile uses 2 registers per thread
tcgen05_ld<num * 2, SHAPE::_16x128b, num>(tmp, row, col);
}
template <int num>
__device__ inline
void tcgen05_ld_16x256b(float *tmp, int row, int col) {
// each 16x256b tile uses 4 registers per thread
tcgen05_ld<num * 4, SHAPE::_16x256b, num>(tmp, row, col);
}
// annoying workaround so that we can modify PTX string from host
enum class L2_MOD { NONE, EVICT_NORMAL, EVICT_FIRST, EVICT_LAST };
template <L2_MOD mod> struct l2_mod_ptx {};
template<> struct l2_mod_ptx<L2_MOD::NONE> { static constexpr char str[] = ""; };
template<> struct l2_mod_ptx<L2_MOD::EVICT_NORMAL> { static constexpr char str[] = ".L2::evict_normal"; };
template<> struct l2_mod_ptx<L2_MOD::EVICT_FIRST> { static constexpr char str[] = ".L2::evict_first"; };
template<> struct l2_mod_ptx<L2_MOD::EVICT_LAST> { static constexpr char str[] = ".L2::evict_last"; };
constexpr int BLOCK_N = 128;
constexpr int BLOCK_K = 256;
constexpr int NUM_WARPS = 6;
constexpr int TB_SIZE = NUM_WARPS * WARP_SIZE;
template <typename T>
__device__ __inline__
T warp_uniform(T x) { return __shfl_sync(0xFFFF'FFFF, x, 0); }
template <int NUM_GROUPS>
struct Arguments {
CUtensorMap A_tmap_list[NUM_GROUPS];
CUtensorMap B_tmap_list[NUM_GROUPS];
CUtensorMap SFA_tmap_list[NUM_GROUPS];
CUtensorMap SFB_tmap_list[NUM_GROUPS];
half *C_ptr_list[NUM_GROUPS];
int M_list[NUM_GROUPS];
int grid_m_cu[NUM_GROUPS + 1];
};
template <L2_MOD l2_mod>
__device__ __inline__
void stg_16(half *ptr, float *tmp) {
asm volatile(
"cvt.rn.f16x2.f32 %0, %1, %0;\n"
"cvt.rn.f16x2.f32 %2, %3, %2;\n"
"cvt.rn.f16x2.f32 %4, %5, %4;\n"
"cvt.rn.f16x2.f32 %6, %7, %6;\n"
"cvt.rn.f16x2.f32 %8, %9, %8;\n"
"cvt.rn.f16x2.f32 %10, %11, %10;\n"
"cvt.rn.f16x2.f32 %12, %13, %12;\n"
"cvt.rn.f16x2.f32 %14, %15, %14;\n"
"st.relaxed.cta.global.L1::no_allocate%17.v8.b32 [%16], {%0, %2, %4, %6, %8, %10, %12, %14};"
: "+f"(tmp[0]), "+f"(tmp[1]), "+f"(tmp[2]), "+f"(tmp[3]),
"+f"(tmp[4]), "+f"(tmp[5]), "+f"(tmp[6]), "+f"(tmp[7]),
"+f"(tmp[8]), "+f"(tmp[9]), "+f"(tmp[10]), "+f"(tmp[11]),
"+f"(tmp[12]), "+f"(tmp[13]), "+f"(tmp[14]), "+f"(tmp[15])
: "l"(ptr), "C"(l2_mod_ptx<l2_mod>::str)
);
}
template <int NUM_GROUPS, int BLOCK_M, int N, int K, int NUM_STAGES, int CTA_GROUP>
__global__
__cluster_dims__(CTA_GROUP, 1, 1)
__launch_bounds__(TB_SIZE)
void kernel_cutlass(const __grid_constant__ Arguments<NUM_GROUPS> args, int K_dyn) {
const int tid = threadIdx.x;
const int lane_id = tid % WARP_SIZE;
const int warp_id = warp_uniform(tid / WARP_SIZE);
// set up smem
extern __shared__ __align__(1024) char smem_ptr[];
const int smem = static_cast<int>(__cvta_generic_to_shared(smem_ptr));
constexpr int A_size = (BLOCK_M / CTA_GROUP) * BLOCK_K / 2;
constexpr int B_size = BLOCK_N * BLOCK_K / 2;
constexpr int SF_size = 128 * BLOCK_K / 16;
constexpr int SFA_size = SF_size * (BLOCK_M / 128);
constexpr int STAGE_SIZE = A_size + B_size + SF_size + SFA_size;
// set up mbarriers and tmem
const int tma_mbar_addr = smem + NUM_STAGES * STAGE_SIZE;
const int mma_mbar_addr = tma_mbar_addr + NUM_STAGES * 8;
const int mainloop_mbar_addr = mma_mbar_addr + NUM_STAGES * 8;
const int epilogue_mbar_addr = mainloop_mbar_addr + 3 * 8;
const int taddr_addr = epilogue_mbar_addr + 3 * 8;
constexpr uint64_t cache_A = EVICT_FIRST;
constexpr uint64_t cache_B = EVICT_FIRST;
constexpr int bar_epilogue = 2;
constexpr int rest_k = K / 16 / 4;
if (warp_id == 0 && elect_sync()) {
// not important that we prefetch tmap for the corresponding GEMM group
int group_id = blockIdx.x % NUM_GROUPS;
prefetch_tensormap(args.A_tmap_list + group_id);
prefetch_tensormap(args.B_tmap_list + group_id);
prefetch_tensormap(args.SFA_tmap_list + group_id);
prefetch_tensormap(args.SFB_tmap_list + group_id);
}
else if (warp_id == 1 && elect_sync()) {
// 1 thread init mbarrier
for (int i = 0; i < NUM_STAGES; i++) {
mbarrier_init(tma_mbar_addr + i * 8, CTA_GROUP);
mbarrier_init(mma_mbar_addr + i * 8, 1);
}
for (int i = 0; i < 3; i++)
mbarrier_init(mainloop_mbar_addr + i * 8, 1);
for (int i = 0; i < 3; i++)
mbarrier_init(epilogue_mbar_addr + i * 8, 4 * WARP_SIZE * CTA_GROUP);
asm volatile("fence.mbarrier_init.release.cluster;"); // visible to async proxy
}
if constexpr (CTA_GROUP == 2) {
asm volatile("barrier.cluster.arrive.relaxed.aligned;");
asm volatile("barrier.cluster.wait.acquire.aligned;");
}
else {
__syncthreads();
}
constexpr int grid_n = N / BLOCK_N;
const int num_m_tiles = args.grid_m_cu[NUM_GROUPS];
const int cta_rank = CTA_GROUP == 2 ? blockIdx.x % CTA_GROUP : 0;
const int bid_n = warp_uniform(blockIdx.x);
// group_id, M, bid_m will be mutated
auto find_bid = [&](int raw_bid_m, int& group_id, int& M, int& bid_m) {
// for loop version
if constexpr (false) {
for (; group_id < NUM_GROUPS; group_id++)
if (raw_bid_m < args.grid_m_cu[group_id + 1]) break;
}
// switch version
// this is faster than for loop version for some reasons.
if constexpr (true) {
switch (group_id) {
case 0: if (raw_bid_m < args.grid_m_cu[1]) break; group_id++;
case 1: if (raw_bid_m < args.grid_m_cu[2]) break; group_id++;
case 2: if (raw_bid_m < args.grid_m_cu[3]) break; group_id++;
case 3: if (raw_bid_m < args.grid_m_cu[4]) break; group_id++;
case 4: if (raw_bid_m < args.grid_m_cu[5]) break; group_id++;
case 5: if (raw_bid_m < args.grid_m_cu[6]) break; group_id++;
case 6: if (raw_bid_m < args.grid_m_cu[7]) break; group_id++;
case 7: if (raw_bid_m < args.grid_m_cu[8]) break; group_id++;
}
}
bid_m = raw_bid_m - args.grid_m_cu[group_id];
M = args.M_list[group_id];
};
if (warp_id == NUM_WARPS - 2) {
// TMA warp
if (elect_sync()) {
int stage_id = 0;
int mma_phase = 0;
const int tma_mbar_addr_ = CTA_GROUP == 2 ? (tma_mbar_addr & 0xFEFFFFFF) : tma_mbar_addr; // report to CTA0
const int off_n = bid_n * BLOCK_N;
const int16_t cta_mask = (1 << CTA_GROUP) - 1;
// variables for closure capture
int raw_bid_m = blockIdx.y;
int group_id = 0;
int M, bid_m, off_m, tma_size;
bool issue_A0, issue_A1, issue_SFA;
const CUtensorMap *A_tmap, *B_tmap, *SFA_tmap, *SFB_tmap;
// consider CTA_GROUP=2 only
// for A, each CTA always issues 64x128B, so 2 CTAs hold 128x128B tile together.
// it means that for BLOCK_M=256, each CTA needs to issue 2 TMA.
// (we do this because we use MMA_N=128 even when BLOCK_M=256)
// for SFA, we can't load the SFA tile corresponding a 64x128B A tile, because of
// the 32x4x4 layout. hence, we do things slightly different from A:
// - BLOCK_M=128: each CTA loads half of 32x4x4 tile, mulicast.
// - BLOCK_M=256: each CTA loads one 32x4x4 tile, multicast.
// do_wait will be inlined
auto issue_tma = [&](int iter_k, int &stage_id, bool do_wait) {
// select tma mbar and smem
const int mbar_addr = tma_mbar_addr_ + stage_id * 8;
const int B_smem = smem + stage_id * STAGE_SIZE;
const int A_smem = B_smem + B_size;
const int SFB_smem = A_smem + A_size;
const int SFA_smem = SFB_smem + SF_size + cta_rank * (SFA_size / 2);
// divide by 8 because we use int64 as dtype for tensor map (to get around boxDim<=256 restriction)
const int off_sfb = (bid_n * rest_k * 512 + iter_k * SF_size) / 8;
const int off_sfa = BLOCK_M == 128
? (bid_m * rest_k * 512 + iter_k * SF_size + cta_rank * (SFA_size / 2)) / 8
: ((bid_m * (BLOCK_M / 128) + cta_rank) * rest_k * 512 + iter_k * SF_size) / 8;
// wait MMA
if (do_wait)
mbarrier_wait(mma_mbar_addr + stage_id * 8, mma_phase);
// issue MMA
tma_3d_g2s<CTA_GROUP>(B_smem, B_tmap, 0, off_n, iter_k, mbar_addr, cache_B);
tma_1d_g2s<CTA_GROUP>(SFB_smem, SFB_tmap, off_sfb, mbar_addr, cache_B);
// NOTE: we will get illegal memory access if off_sfa is out-of-bounds.
if (issue_A0) tma_3d_g2s<CTA_GROUP>(A_smem, A_tmap, 0, off_m, iter_k, mbar_addr, cache_A);
if (issue_A1) tma_3d_g2s<CTA_GROUP>(A_smem + A_size / 2, A_tmap, 0, off_m + 128, iter_k, mbar_addr, cache_A);
if (issue_SFA) tma_1d_g2s_mcast<CTA_GROUP>(SFA_smem, SFA_tmap, off_sfa, mbar_addr, cta_mask, cache_A);
// signal TMA done
mbarrier_arrive_expect_tx(mbar_addr, tma_size);
if (do_wait) {
stage_id = (stage_id + 1) % NUM_STAGES;
if (stage_id == 0)
mma_phase ^= 1;
}
};
// unroll the 1st iteration, which skips MMA wait
// this is incorrect if we have (1) iters_k < NUM_STAGES, and (2) more than 1 wave.
// none of the benchmark shapes have these properties, so we don't need to handle it here.
{
find_bid(raw_bid_m, group_id, M, bid_m);
off_m = bid_m * BLOCK_M + cta_rank * (128 / CTA_GROUP);
issue_A0 = off_m < M;
issue_A1 = BLOCK_M == 256 && off_m + 128 < M;
issue_SFA = BLOCK_M == 128 || (bid_m * BLOCK_M + cta_rank * 128 < M); // when BLOCK_M=128, we must always issue SFA TMA
tma_size = B_size + SF_size + (issue_SFA ? SFA_size : 0);
if (issue_A0) tma_size += (128 / CTA_GROUP) * BLOCK_K / 2;
if (issue_A1) tma_size += (128 / CTA_GROUP) * BLOCK_K / 2;
A_tmap = args.A_tmap_list + group_id;
B_tmap = args.B_tmap_list + group_id;
SFA_tmap = args.SFA_tmap_list + group_id;
SFB_tmap = args.SFB_tmap_list + group_id;
#pragma unroll 1
for (int iter_k = 0; iter_k < std::min(NUM_STAGES, K / BLOCK_K); iter_k++)
issue_tma(iter_k, iter_k, false);
// the rest of the 1st wave
for (int iter_k = NUM_STAGES; iter_k < K_dyn / BLOCK_K; iter_k++)
issue_tma(iter_k, stage_id, true);
raw_bid_m += gridDim.y;
}
for (; raw_bid_m < num_m_tiles; raw_bid_m += gridDim.y) {
find_bid(raw_bid_m, group_id, M, bid_m);
off_m = bid_m * BLOCK_M + cta_rank * (128 / CTA_GROUP);
issue_A0 = off_m < M;
issue_A1 = BLOCK_M == 256 && off_m + 128 < M;
issue_SFA = BLOCK_M == 128 || (bid_m * BLOCK_M + cta_rank * 128 < M); // when BLOCK_M=128, we must always issue SFA TMA
tma_size = B_size + SF_size + (issue_SFA ? SFA_size : 0);
if (issue_A0) tma_size += (128 / CTA_GROUP) * BLOCK_K / 2;
if (issue_A1) tma_size += (128 / CTA_GROUP) * BLOCK_K / 2;
A_tmap = args.A_tmap_list + group_id;
B_tmap = args.B_tmap_list + group_id;
SFA_tmap = args.SFA_tmap_list + group_id;
SFB_tmap = args.SFB_tmap_list + group_id;
for (int iter_k = 0; iter_k < K_dyn / BLOCK_K; iter_k++)
issue_tma(iter_k, stage_id, true);
}
}
}
else if (warp_id == NUM_WARPS - 1) {
// MMA warp
tcgen05_alloc<CTA_GROUP>(taddr_addr, 512); // allocate tmem
// instruction desc
// always use MMA_N=128 regardless of BLOCK_M value
constexpr uint32_t MMA_M = BLOCK_N * CTA_GROUP;
constexpr uint32_t MMA_N = 128;
constexpr uint32_t i_desc = (1U << 7U) // atype=E2M1
| (1U << 10U) // btype=E2M1
| (MMA_N >> 3U << 17U)
| (MMA_M >> 7U << 27U)
;
if (cta_rank == 0 && elect_sync()) {
int outer_stage0 = 0;
int outer_stage1 = 1;
// used by BLOCK_M=128 and BLOCK_M=256 respectively.
// the compiler will remove the unused one.
// we used shared memory for BLOCK_M=256 because we need dynamic indexing.
int epilogue_phase_128 = 1;
int epilogue_phase_256[3] = {1, 1, 1};
int inner_stage = 0;
int tma_phase = 0;
// for BLOCK_M=256, we use 3 tmem buffers
// 1st wave: buffer0 and buffer1. epilogue loads buffer1 first.
// 2nd wave: buffer1 and buffer2. we can start buffer2 first. then wait for buffer1
const int16_t cta_mask = (1 << CTA_GROUP) - 1;
int group_id = 0;
int M, bid_m;
bool do_2nd_mma = false;
for (int raw_bid_m = blockIdx.y; raw_bid_m < num_m_tiles; raw_bid_m += gridDim.y) {
// we can skip the 2nd MMA
if constexpr (BLOCK_M == 256) {
find_bid(raw_bid_m, group_id, M, bid_m);
do_2nd_mma = bid_m * BLOCK_M + 128 < M;
}
const int acc0_tmem = outer_stage0 * 128;
const int acc1_tmem = outer_stage1 * 128;
// do_wait and do_commit will be inlined
auto issue_mma = [&](int enable_input_d, bool do_wait, bool do_commit) {
// wait for the 1st buffer
if (do_wait) {
if constexpr (BLOCK_M == 128) {
mbarrier_wait(epilogue_mbar_addr + outer_stage0 * 8, epilogue_phase_128);
}
if constexpr (BLOCK_M == 256) {
// do this to avoid local memory (due to dynamic indexing)
if (outer_stage0 == 0) {
mbarrier_wait(epilogue_mbar_addr + 0 * 8, epilogue_phase_256[0]);
epilogue_phase_256[0] ^= 1;
}
else if (outer_stage0 == 1) {
mbarrier_wait(epilogue_mbar_addr + 1 * 8, epilogue_phase_256[1]);
epilogue_phase_256[1] ^= 1;
}
else {
mbarrier_wait(epilogue_mbar_addr + 2 * 8, epilogue_phase_256[2]);
epilogue_phase_256[2] ^= 1;
}
}
}
// select smem
const int B_smem = smem + inner_stage * STAGE_SIZE;
const int A0_smem = B_smem + B_size;
const int A1_smem = A0_smem + A_size / 2;
const int SFB_smem = A1_smem + A_size / 2;
const int SFA0_smem = SFB_smem + SF_size;
const int SFA1_smem = SFA0_smem + SF_size;
// set up smem desc
// AB: 128-byte swizzling
constexpr uint64_t AB_desc = (desc_encode(8 * 128) << 32ULL) | (1ULL << 46ULL) | (2ULL << 61ULL);
uint64_t b_desc = AB_desc | (B_smem >> 4);
uint64_t a0_desc = AB_desc | (A0_smem >> 4);
uint64_t a1_desc = AB_desc | (A1_smem >> 4);
// SF: no swizzling
constexpr uint64_t SF_desc = (desc_encode(8 * 16) << 32ULL) | (1ULL << 46ULL);
uint64_t sfb_desc = SF_desc | (SFB_smem >> 4);
uint64_t sfa0_desc = SF_desc | (SFA0_smem >> 4);
uint64_t sfa1_desc = SF_desc | (SFA1_smem >> 4);
// each SF consumes 16 columns per BLOCK_K=256
int sfb_tmem = 128 * 3;
int sfa0_tmem = sfb_tmem + 16;
int sfa1_tmem = sfa0_tmem + 4;
// wait TMA
mbarrier_wait(tma_mbar_addr + inner_stage * 8, tma_phase);
// manual unroll 1st iteration
tcgen05_cp_nvfp4<CTA_GROUP>(sfb_tmem, sfb_desc);
tcgen05_cp_nvfp4<CTA_GROUP>(sfa0_tmem, sfa0_desc);
tcgen05_mma_nvfp4<CTA_GROUP>(acc0_tmem, b_desc, a0_desc, i_desc, sfb_tmem, sfa0_tmem, enable_input_d);
for (int k = 1; k < BLOCK_K / MMA_K; k++) {
// next 4 columns
sfb_tmem += 4;
sfa0_tmem += 4 * (BLOCK_M / 128);
// next 512-byte
sfb_desc += (512 >> 4);
sfa0_desc += (512 >> 4);
// next 32-byte
b_desc += (32 >> 4);
a0_desc += (32 >> 4);
tcgen05_cp_nvfp4<CTA_GROUP>(sfb_tmem, sfb_desc);
tcgen05_cp_nvfp4<CTA_GROUP>(sfa0_tmem, sfa0_desc);
tcgen05_mma_nvfp4<CTA_GROUP>(acc0_tmem, b_desc, a0_desc, i_desc, sfb_tmem, sfa0_tmem, 1);
}
// signal mainloop done
if (do_commit)
tcgen05_commit_mcast<CTA_GROUP>(mainloop_mbar_addr + outer_stage0 * 8, cta_mask);
if (BLOCK_M == 256 && do_2nd_mma) {
// wait for the 2nd buffer
if (do_wait) {
if (outer_stage1 == 0) {
mbarrier_wait(epilogue_mbar_addr + 0 * 8, epilogue_phase_256[0]);
epilogue_phase_256[0] ^= 1;
}
else if (outer_stage1 == 1) {
mbarrier_wait(epilogue_mbar_addr + 1 * 8, epilogue_phase_256[1]);
epilogue_phase_256[1] ^= 1;
}
else {
mbarrier_wait(epilogue_mbar_addr + 2 * 8, epilogue_phase_256[2]);
epilogue_phase_256[2] ^= 1;
}
}
uint64_t b_desc = AB_desc | (B_smem >> 4);
int sfb_tmem = 128 * 3;
tcgen05_cp_nvfp4<CTA_GROUP>(sfa1_tmem, sfa1_desc);
tcgen05_mma_nvfp4<CTA_GROUP>(acc1_tmem, b_desc, a1_desc, i_desc, sfb_tmem, sfa1_tmem, enable_input_d);
for (int k = 1; k < BLOCK_K / MMA_K; k++) {
// next 4 columns
sfb_tmem += 4;
sfa1_tmem += 4 * (BLOCK_M / 128);
// next 512-byte
sfa1_desc += (512 >> 4);
// next 32-byte
b_desc += (32 >> 4);
a1_desc += (32 >> 4);
tcgen05_cp_nvfp4<CTA_GROUP>(sfa1_tmem, sfa1_desc);
tcgen05_mma_nvfp4<CTA_GROUP>(acc1_tmem, b_desc, a1_desc, i_desc, sfb_tmem, sfa1_tmem, 1);
}
// signal mainloop done
if (do_commit)
tcgen05_commit_mcast<CTA_GROUP>(mainloop_mbar_addr + outer_stage1 * 8, cta_mask);
}
// signal MMA done
tcgen05_commit_mcast<CTA_GROUP>(mma_mbar_addr + inner_stage * 8, cta_mask);
inner_stage = (inner_stage + 1) % NUM_STAGES;
if (inner_stage == 0)
tma_phase ^= 1;
};
// unroll the 1st iteration to wait for each buffer separately
issue_mma(0, true, false);
// we use K_dyn to prevent the compiler from unrolling this loop.
// when using cutlass incantation, adding #pragma unroll 1 to this loop
// results in segmentation fault.
for (int iter_k = 1; iter_k < K_dyn / BLOCK_K - 1; iter_k++)
issue_mma(1, false, false);
// unroll the last iteration to commit each buffer separately
issue_mma(1, false, true);
if constexpr (BLOCK_M == 128) {
outer_stage0 ^= 1;
if (outer_stage0 == 0)
epilogue_phase_128 ^= 1;
}
if constexpr (BLOCK_M == 256) {
outer_stage0 = (outer_stage0 + 2) % 3;
outer_stage1 = (outer_stage1 + 2) % 3;
}
}
}
}
else {
// epilogue warps
int stage0 = 0;
int stage1 = 1;
int mainloop_phase_128 = 0;
int mainloop_phase_256[3] = {0, 0, 0};
const int epilogue_mbar_addr_ = CTA_GROUP == 2 ? (epilogue_mbar_addr & 0xFEFFFFFF) : epilogue_mbar_addr; // report to CTA0
const int off_n = bid_n * BLOCK_N;
const int row = off_n + tid;
int raw_bid_m = blockIdx.y;
int group_id = 0;
int M, bid_m;
bool do_2nd_mma = false;
float tmp[128];
// unroll the last iteration
for (; raw_bid_m < num_m_tiles - gridDim.y; raw_bid_m += gridDim.y) {
find_bid(raw_bid_m, group_id, M, bid_m);
const int off_m = bid_m * BLOCK_M;
half *C_ptr = args.C_ptr_list[group_id];
const int stride_cn = cdiv(M, 16) * 16; // multiple of 16
if constexpr (BLOCK_M == 256) {
find_bid(raw_bid_m, group_id, M, bid_m);
do_2nd_mma = bid_m * BLOCK_M + 128 < M;
}
// stage0
if (warp_id == 0) {
if constexpr (BLOCK_M == 128)
mbarrier_wait(mainloop_mbar_addr + stage0 * 8, mainloop_phase_128);
if constexpr (BLOCK_M == 256) {
if (stage0 == 0) {
mbarrier_wait(mainloop_mbar_addr + 0 * 8, mainloop_phase_256[0]);
mainloop_phase_256[0] ^= 1;
}
else if (stage0 == 1) {
mbarrier_wait(mainloop_mbar_addr + 1 * 8, mainloop_phase_256[1]);
mainloop_phase_256[1] ^= 1;
}
else {
mbarrier_wait(mainloop_mbar_addr + 2 * 8, mainloop_phase_256[2]);
mainloop_phase_256[2] ^= 1;
}
}
}
bar_sync<bar_epilogue>(4 * WARP_SIZE);
asm volatile("tcgen05.fence::after_thread_sync;");
tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage0 * 128);
asm volatile("tcgen05.wait::ld.sync.aligned;");
mbarrier_arrive(epilogue_mbar_addr_ + stage0 * 8);
for (int m = 0; m < 128 / 16; m++) {
const int col = off_m + m * 16;
if (col >= M) break;
stg_16<L2_MOD::EVICT_LAST>(C_ptr + (row * stride_cn + col), tmp + m * 16);
}
if constexpr (BLOCK_M == 256) {
if (do_2nd_mma) {
// stage1
if (warp_id == 0) {
if (stage1 == 0) {
mbarrier_wait(mainloop_mbar_addr + 0 * 8, mainloop_phase_256[0]);
mainloop_phase_256[0] ^= 1;
}
else if (stage1 == 1) {
mbarrier_wait(mainloop_mbar_addr + 1 * 8, mainloop_phase_256[1]);
mainloop_phase_256[1] ^= 1;
}
else {
mbarrier_wait(mainloop_mbar_addr + 2 * 8, mainloop_phase_256[2]);
mainloop_phase_256[2] ^= 1;
}
}
bar_sync<bar_epilogue>(4 * WARP_SIZE);
asm volatile("tcgen05.fence::after_thread_sync;");
tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage1 * 128);
asm volatile("tcgen05.wait::ld.sync.aligned;");
mbarrier_arrive(epilogue_mbar_addr_ + stage1 * 8);
for (int m = 0; m < 128 / 16; m++) {
const int col = off_m + 128 + m * 16;
if (col >= M) break;
stg_16<L2_MOD::EVICT_LAST>(C_ptr + (row * stride_cn + col), tmp + m * 16);
}
}
else {
// arrive immediately
mbarrier_arrive(epilogue_mbar_addr_ + stage1 * 8);
}
}
if constexpr (BLOCK_M == 128) {
stage0 ^= 1;
if (stage0 == 0)
mainloop_phase_128 ^= 1;
}
if constexpr (BLOCK_M == 256) {
stage0 = (stage0 + 2) % 3;
stage1 = (stage1 + 2) % 3;
}
}
{
find_bid(raw_bid_m, group_id, M, bid_m);
const int off_m = bid_m * BLOCK_M;
half *C_ptr = args.C_ptr_list[group_id];
const int stride_cn = cdiv(M, 16) * 16; // multiple of 16
if constexpr (BLOCK_M == 256) {
find_bid(raw_bid_m, group_id, M, bid_m);
do_2nd_mma = bid_m * BLOCK_M + 128 < M;
}
// stage0
if (warp_id == 0) {
if constexpr (BLOCK_M == 128)
mbarrier_wait(mainloop_mbar_addr + stage0 * 8, mainloop_phase_128);
if constexpr (BLOCK_M == 256) {
if (stage0 == 0) {
mbarrier_wait(mainloop_mbar_addr + 0 * 8, mainloop_phase_256[0]);
mainloop_phase_256[0] ^= 1;
}
else if (stage0 == 1) {
mbarrier_wait(mainloop_mbar_addr + 1 * 8, mainloop_phase_256[1]);
mainloop_phase_256[1] ^= 1;
}
else {
mbarrier_wait(mainloop_mbar_addr + 2 * 8, mainloop_phase_256[2]);
mainloop_phase_256[2] ^= 1;
}
}
}
bar_sync<bar_epilogue>(4 * WARP_SIZE);
asm volatile("tcgen05.fence::after_thread_sync;");
tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage0 * 128);
asm volatile("tcgen05.wait::ld.sync.aligned;");
for (int m = 0; m < 128 / 16; m++) {
const int col = off_m + m * 16;
if (col >= M) break;
stg_16<L2_MOD::NONE>(C_ptr + (row * stride_cn + col), tmp + m * 16);
}
if constexpr (BLOCK_M == 256) {
if (do_2nd_mma) {
// stage1
if (warp_id == 0) {
if (stage1 == 0)
mbarrier_wait(mainloop_mbar_addr + 0 * 8, mainloop_phase_256[0]);
else if (stage1 == 1)
mbarrier_wait(mainloop_mbar_addr + 1 * 8, mainloop_phase_256[1]);
else
mbarrier_wait(mainloop_mbar_addr + 2 * 8, mainloop_phase_256[2]);
}
bar_sync<bar_epilogue>(4 * WARP_SIZE);
asm volatile("tcgen05.fence::after_thread_sync;");
tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage1 * 128);
asm volatile("tcgen05.wait::ld.sync.aligned;");
for (int m = 0; m < 128 / 16; m++) {
const int col = off_m + 128 + m * 16;
if (col >= M) break;
stg_16<L2_MOD::NONE>(C_ptr + (row * stride_cn + col), tmp + m * 16);
}
}
}
}
if constexpr (CTA_GROUP == 2) {
asm volatile("barrier.cluster.arrive.relaxed.aligned;");
if (warp_id == 0) {
asm volatile("barrier.cluster.wait.acquire.aligned;");
tcgen05_dealloc<CTA_GROUP>(0, 512);
}
}
if constexpr (CTA_GROUP == 1) {
bar_sync<bar_epilogue>(4 * WARP_SIZE);
if (warp_id == 0)
tcgen05_dealloc<CTA_GROUP>(0, 512);
}
}
}
void check_cu(CUresult err) {
if (err == CUDA_SUCCESS) return;
const char *error_msg_ptr;
if (cuGetErrorString(err, &error_msg_ptr) != CUDA_SUCCESS)
error_msg_ptr = "unable to get error string";
TORCH_CHECK(false, "cuTensorMapEncodeTiled error: ", error_msg_ptr);
}
void check_cuda(cudaError_t err) {
if (err == cudaSuccess) return;
TORCH_CHECK(false, cudaGetErrorString(err));
}
void init_AB_tmap(
CUtensorMap *tmap,
void *ptr,
uint64_t global_height, uint64_t global_width,
uint32_t shared_height, uint32_t shared_width
) {
constexpr uint32_t rank = 3;
uint64_t globalDim[rank] = {256, global_height, global_width / 256};
uint64_t globalStrides[rank-1] = {global_width / 2, 128}; // in bytes
uint32_t boxDim[rank] = {256, shared_height, shared_width / 256};
uint32_t elementStrides[rank] = {1, 1, 1};
auto err = cuTensorMapEncodeTiled(
tmap,
CUtensorMapDataType::CU_TENSOR_MAP_DATA_TYPE_16U4_ALIGN8B,
rank,
ptr,
globalDim,
globalStrides,
boxDim,
elementStrides,
CUtensorMapInterleave::CU_TENSOR_MAP_INTERLEAVE_NONE,
CUtensorMapSwizzle::CU_TENSOR_MAP_SWIZZLE_128B,
CUtensorMapL2promotion::CU_TENSOR_MAP_L2_PROMOTION_NONE,
CUtensorMapFloatOOBfill::CU_TENSOR_MAP_FLOAT_OOB_FILL_NONE
);
//check_cu(err);
}
void init_SF_tmap(CUtensorMap *tmap, void *ptr, uint64_t MN, uint64_t K, uint32_t SF_size) {
MN = cdiv(MN, 128) * 128; // round up to multiple of 128
const uint64_t global_size = MN * K / 16;
// use int64 as dtype, hence divide sizes by 8
constexpr uint32_t rank = 1;
uint64_t globalDim[rank] = {global_size / 8};
uint64_t globalStrides[rank-1] = {}; // in bytes
uint32_t boxDim[rank] = {SF_size / 8};
uint32_t elementStrides[rank] = {1};
auto err = cuTensorMapEncodeTiled(
tmap,
CUtensorMapDataType::CU_TENSOR_MAP_DATA_TYPE_INT64,
rank,
ptr,
globalDim,
globalStrides,
boxDim,
elementStrides,
CUtensorMapInterleave::CU_TENSOR_MAP_INTERLEAVE_NONE,
CUtensorMapSwizzle::CU_TENSOR_MAP_SWIZZLE_NONE,
CUtensorMapL2promotion::CU_TENSOR_MAP_L2_PROMOTION_NONE,
CUtensorMapFloatOOBfill::CU_TENSOR_MAP_FLOAT_OOB_FILL_NONE
);
//check_cu(err);
}
template <int NUM_GROUPS, int BLOCK_M, int N, int K, int CTA_GROUP>
Arguments<NUM_GROUPS> create_args_template() {
Arguments<NUM_GROUPS> args;
args.grid_m_cu[0] = 0;
for (int i = 0; i < NUM_GROUPS; i++) {
init_AB_tmap(args.A_tmap_list + i, nullptr, 1, K, 128 / CTA_GROUP, BLOCK_K);
init_AB_tmap(args.B_tmap_list + i, nullptr, N, K, BLOCK_N, BLOCK_K);
init_SF_tmap(args.SFA_tmap_list + i, nullptr, 1, K, BLOCK_M * BLOCK_K / 16 / CTA_GROUP);
init_SF_tmap(args.SFB_tmap_list + i, nullptr, N, K, BLOCK_N * BLOCK_K / 16);
}
return args;
}
// from ChatGPT
template <int N>
void argsort_desc(const int (&values)[N], int (&indices)[N]) {
// initialize indices
for (int i = 0; i < N; ++i)
indices[i] = i;
// sort indices by values
std::sort(indices, indices + N, [&](int i, int j) { return values[i] > values[j]; });
}
template <int NUM_GROUPS, int BLOCK_M, int N, int K, int CTA_GROUP>
void group_gemm_launch(
at::TensorList A_list,
at::TensorList B_list,
at::TensorList SFA_list,
at::TensorList SFB_list,
at::TensorList C_list
) {
constexpr int grid_n = N / BLOCK_N;
// notice static. we init once, then only change M and pointer addresses.
static Arguments<NUM_GROUPS> args = create_args_template<NUM_GROUPS, BLOCK_M, N, K, CTA_GROUP>();
// sort by descending M values
// this helps benchmark.0 a bit, probably thanks to reduced tail effect of the epilogue.
int values[NUM_GROUPS];
for (int i = 0; i < NUM_GROUPS; i++) {
values[i] = A_list[i].size(0);
}
int indices[NUM_GROUPS];
argsort_desc<NUM_GROUPS>(values, indices);
for (int i = 0; i < NUM_GROUPS; i++) {
const int idx = indices[i];
const int M = A_list[idx].size(0);
// exploit the internal encodings of CUtensorMap. doesn't seem to be faster.
reinterpret_cast<void **>(args.A_tmap_list + i)[0] = A_list[idx].data_ptr();
reinterpret_cast<int *>(args.A_tmap_list + i)[9] = M - 1;
reinterpret_cast<void **>(args.B_tmap_list + i)[0] = B_list[idx].data_ptr();
reinterpret_cast<void **>(args.SFA_tmap_list + i)[0] = SFA_list[idx].data_ptr();
reinterpret_cast<int *>(args.SFA_tmap_list + i)[8] = (cdiv(M, 128) * 128 * K / 16) - 1;
reinterpret_cast<void **>(args.SFB_tmap_list + i)[0] = SFB_list[idx].data_ptr();
args.C_ptr_list[i] = reinterpret_cast<half *>(C_list[idx].data_ptr());
args.M_list[i] = M;
args.grid_m_cu[i + 1] = args.grid_m_cu[i] + cdiv(M, BLOCK_M);
}
// make sure num SMs used is a multiple of grid_n
const dim3 grid(grid_n, std::min(148 / grid_n, args.grid_m_cu[NUM_GROUPS]));
constexpr int AB_size = ((BLOCK_M / CTA_GROUP) + BLOCK_N) * (BLOCK_K / 2);
constexpr int SF_size = 128 * (BLOCK_K / 16) * (1 + BLOCK_M / 128);
constexpr int sm100_size = 227 * 1024;
constexpr int dynamic_size = AB_size + SF_size + 2 * 8; // 1 tma_mbar, 1 mma_mbar
constexpr int static_size = 3 * 2 * 8 + 4; // 3 mainloop_mbar, 3 epilogue_mbar, tmem_addr
constexpr int NUM_STAGES = (sm100_size - static_size) / dynamic_size;
constexpr int smem_size = dynamic_size * NUM_STAGES + static_size;
// cutlass incantation (this affects ptxas)
auto this_kernel = kernel_cutlass<NUM_GROUPS, BLOCK_M, N, K, NUM_STAGES, CTA_GROUP>;
cudaFuncSetAttribute(this_kernel, cudaFuncAttributeMaxDynamicSharedMemorySize, smem_size);
this_kernel<<<grid, TB_SIZE, smem_size>>>(args, K);
}
void group_gemm(
at::TensorList A_list,
at::TensorList B_list,
at::TensorList SFA_list,
at::TensorList SFB_list,
at::TensorList C_list
) {
const int G = A_list.size();
const int N = B_list[0].size(0);
const int K = B_list[0].size(1) * 2;
#define LAUNCH(G_, BM, N_, K_, CTA_GROUP) \
else if (G == G_ && N == N_ && K == K_) { \
group_gemm_launch<G_, BM, N_, K_, CTA_GROUP>(A_list, B_list, SFA_list, SFB_list, C_list); \
}
if (false) {}
LAUNCH(8, 128, 4096, 7168, 2)
LAUNCH(8, 256, 7168, 2048, 2)
LAUNCH(2, 128, 3072, 4096, 2)
LAUNCH(2, 128, 4096, 1536, 2)
#undef LAUNCH
}
TORCH_LIBRARY(my_module, m) {
m.def("group_gemm(Tensor[] A_list, Tensor[] B_list, Tensor[] SFA_list, Tensor[] SFB_list, Tensor(a!)[] C_list) -> ()");
m.impl("group_gemm", &group_gemm);
}
"""
load_inline(
"group_gemm",
cpp_sources="",
cuda_sources=CUDA_SRC,
verbose=True,
is_python_module=False,
no_implicit_headers=True,
extra_cuda_cflags=[
"-O3",
"-gencode=arch=compute_100a,code=sm_100a",
"--use_fast_math",
"--expt-relaxed-constexpr",
"--relocatable-device-code=false",
"-lineinfo",
"-Xptxas=-v",
# "--keep",
# "--keep-dir",
# f"{Path(__file__).parent}/tmp",
],
extra_ldflags=["-lcuda"],
)
group_gemm = torch.ops.my_module.group_gemm
def ref(A_list, B_list, SFA_list, SFB_list, C_list):
for a, b, sfa, sfb, c in zip(A_list, B_list, SFA_list, SFB_list, C_list):
torch._scaled_mm(
a[..., 0],
b[..., 0].T,
sfa.permute(5, 2, 4, 0, 1, 3).view(-1),
sfb.permute(5, 2, 4, 0, 1, 3).view(-1),
out=c[..., 0],
)
def custom_kernel(data: input_t) -> output_t:
abc_list, _, sf_list, shape_list = data
A_list, B_list, C_list = zip(*abc_list)
SFA_list, SFB_list = zip(*sf_list)
_, N0, K0, _ = shape_list[0]
# M-major, and pad M to multiple of 16
C_list = []
for M, N, _, _ in shape_list:
new_M = (M + 16 - 1) // 16 * 16
new_C = torch.empty(new_M * N, dtype=torch.half, device="cuda")
new_C = new_C.as_strided((M, N, 1), (1, new_M, 0))
C_list.append(new_C)
for _, N, K, _ in shape_list:
if N != N0 or K != K0:
ref(A_list, B_list, SFA_list, SFB_list, C_list)
break
else:
# benchmark shapes: same N and K across groups
group_gemm(A_list, B_list, SFA_list, SFB_list, C_list)
# torch.cuda.synchronize()
return C_list
scrolls · 1262 lines total
Source code from GPU Mode and the KernelBot dataset · June 9 Researcher Reciprocity License v1.0
Changes from previous submission
Against this author's previous submission submission 503798.
⋯ 427 unchanged linesconst int tma_mbar_addr = smem + NUM_STAGES * STAGE_SIZE;const int mma_mbar_addr = tma_mbar_addr + NUM_STAGES * 8;const int mainloop_mbar_addr = mma_mbar_addr + NUM_STAGES * 8;- const int epilogue_mbar_addr = mainloop_mbar_addr + 2 * 8;- const int taddr_addr = epilogue_mbar_addr + 2 * (BLOCK_M / 128) * 8;+ const int epilogue_mbar_addr = mainloop_mbar_addr + 3 * 8;+ const int taddr_addr = epilogue_mbar_addr + 3 * 8;constexpr uint64_t cache_A = EVICT_FIRST;constexpr uint64_t cache_B = EVICT_FIRST;⋯ 15 unchanged linesmbarrier_init(tma_mbar_addr + i * 8, CTA_GROUP);mbarrier_init(mma_mbar_addr + i * 8, 1);}- for (int i = 0; i < 2; i++)+ for (int i = 0; i < 3; i++)mbarrier_init(mainloop_mbar_addr + i * 8, 1);- for (int i = 0; i < 2 * (BLOCK_M / 128); i++)+ for (int i = 0; i < 3; i++)mbarrier_init(epilogue_mbar_addr + i * 8, 4 * WARP_SIZE * CTA_GROUP);asm volatile("fence.mbarrier_init.release.cluster;"); // visible to async proxy}⋯ 167 unchanged lines;if (cta_rank == 0 && elect_sync()) {- int outer_stage = 0;- int epilogue_phase = 1;+ int outer_stage0 = 0;+ int outer_stage1 = 1;+ // used by BLOCK_M=128 and BLOCK_M=256 respectively.+ // the compiler will remove the unused one.+ // we used shared memory for BLOCK_M=256 because we need dynamic indexing.+ int epilogue_phase_128 = 1;+ int epilogue_phase_256[3] = {1, 1, 1};+int inner_stage = 0;int tma_phase = 0;+ // for BLOCK_M=256, we use 3 tmem buffers+ // 1st wave: buffer0 and buffer1. epilogue loads buffer1 first.+ // 2nd wave: buffer1 and buffer2. we can start buffer2 first. then wait for buffer1+const int16_t cta_mask = (1 << CTA_GROUP) - 1;int group_id = 0;int M, bid_m;- bool do_2nd_mma = true;+ bool do_2nd_mma = false;for (int raw_bid_m = blockIdx.y; raw_bid_m < num_m_tiles; raw_bid_m += gridDim.y) {// we can skip the 2nd MMA- if constexpr(BLOCK_M == 256) {+ if constexpr (BLOCK_M == 256) {find_bid(raw_bid_m, group_id, M, bid_m);do_2nd_mma = bid_m * BLOCK_M + 128 < M;}- const int acc0_tmem = outer_stage * 128;- const int acc1_tmem = acc0_tmem + 128;- mbarrier_wait(epilogue_mbar_addr + outer_stage * 8, epilogue_phase);+ const int acc0_tmem = outer_stage0 * 128;+ const int acc1_tmem = outer_stage1 * 128;- // we use K_dyn to prevent the compiler from unrolling this loop.- // when using cutlass incantation, adding #pragma unroll 1 to this loop- // results in segmentation fault.- for (int iter_k = 0; iter_k < K_dyn / BLOCK_K; iter_k++) {+ // do_wait and do_commit will be inlined+ auto issue_mma = [&](int enable_input_d, bool do_wait, bool do_commit) {+ // wait for the 1st buffer+ if (do_wait) {+ if constexpr (BLOCK_M == 128) {+ mbarrier_wait(epilogue_mbar_addr + outer_stage0 * 8, epilogue_phase_128);+ }+ if constexpr (BLOCK_M == 256) {+ // do this to avoid local memory (due to dynamic indexing)+ if (outer_stage0 == 0) {+ mbarrier_wait(epilogue_mbar_addr + 0 * 8, epilogue_phase_256[0]);+ epilogue_phase_256[0] ^= 1;+ }+ else if (outer_stage0 == 1) {+ mbarrier_wait(epilogue_mbar_addr + 1 * 8, epilogue_phase_256[1]);+ epilogue_phase_256[1] ^= 1;+ }+ else {+ mbarrier_wait(epilogue_mbar_addr + 2 * 8, epilogue_phase_256[2]);+ epilogue_phase_256[2] ^= 1;+ }+ }+ }+// select smemconst int B_smem = smem + inner_stage * STAGE_SIZE;const int A0_smem = B_smem + B_size;⋯ 26 unchanged lines// manual unroll 1st iterationtcgen05_cp_nvfp4<CTA_GROUP>(sfb_tmem, sfb_desc);tcgen05_cp_nvfp4<CTA_GROUP>(sfa0_tmem, sfa0_desc);- tcgen05_mma_nvfp4<CTA_GROUP, COLLECTOR_USAGE::A_FILL>(acc0_tmem, b_desc, a0_desc, i_desc, sfb_tmem, sfa0_tmem, iter_k);- if (BLOCK_M == 256 && do_2nd_mma) {- tcgen05_cp_nvfp4<CTA_GROUP>(sfa1_tmem, sfa1_desc);- tcgen05_mma_nvfp4<CTA_GROUP, COLLECTOR_USAGE::A_LASTUSE>(acc1_tmem, b_desc, a1_desc, i_desc, sfb_tmem, sfa1_tmem, iter_k);- }+ tcgen05_mma_nvfp4<CTA_GROUP>(acc0_tmem, b_desc, a0_desc, i_desc, sfb_tmem, sfa0_tmem, enable_input_d);for (int k = 1; k < BLOCK_K / MMA_K; k++) {// next 4 columnssfb_tmem += 4;sfa0_tmem += 4 * (BLOCK_M / 128);- sfa1_tmem += 4 * (BLOCK_M / 128);// next 512-bytesfb_desc += (512 >> 4);sfa0_desc += (512 >> 4);- sfa1_desc += (512 >> 4);// next 32-byteb_desc += (32 >> 4);a0_desc += (32 >> 4);- a1_desc += (32 >> 4);tcgen05_cp_nvfp4<CTA_GROUP>(sfb_tmem, sfb_desc);tcgen05_cp_nvfp4<CTA_GROUP>(sfa0_tmem, sfa0_desc);- tcgen05_mma_nvfp4<CTA_GROUP, COLLECTOR_USAGE::A_FILL>(acc0_tmem, b_desc, a0_desc, i_desc, sfb_tmem, sfa0_tmem, 1);- if (BLOCK_M == 256 && do_2nd_mma) {+ tcgen05_mma_nvfp4<CTA_GROUP>(acc0_tmem, b_desc, a0_desc, i_desc, sfb_tmem, sfa0_tmem, 1);+ }++ // signal mainloop done+ if (do_commit)+ tcgen05_commit_mcast<CTA_GROUP>(mainloop_mbar_addr + outer_stage0 * 8, cta_mask);++ if (BLOCK_M == 256 && do_2nd_mma) {+ // wait for the 2nd buffer+ if (do_wait) {+ if (outer_stage1 == 0) {+ mbarrier_wait(epilogue_mbar_addr + 0 * 8, epilogue_phase_256[0]);+ epilogue_phase_256[0] ^= 1;+ }+ else if (outer_stage1 == 1) {+ mbarrier_wait(epilogue_mbar_addr + 1 * 8, epilogue_phase_256[1]);+ epilogue_phase_256[1] ^= 1;+ }+ else {+ mbarrier_wait(epilogue_mbar_addr + 2 * 8, epilogue_phase_256[2]);+ epilogue_phase_256[2] ^= 1;+ }+ }++ uint64_t b_desc = AB_desc | (B_smem >> 4);+ int sfb_tmem = 128 * 3;++ tcgen05_cp_nvfp4<CTA_GROUP>(sfa1_tmem, sfa1_desc);+ tcgen05_mma_nvfp4<CTA_GROUP>(acc1_tmem, b_desc, a1_desc, i_desc, sfb_tmem, sfa1_tmem, enable_input_d);++ for (int k = 1; k < BLOCK_K / MMA_K; k++) {+ // next 4 columns+ sfb_tmem += 4;+ sfa1_tmem += 4 * (BLOCK_M / 128);++ // next 512-byte+ sfa1_desc += (512 >> 4);++ // next 32-byte+ b_desc += (32 >> 4);+ a1_desc += (32 >> 4);+tcgen05_cp_nvfp4<CTA_GROUP>(sfa1_tmem, sfa1_desc);- tcgen05_mma_nvfp4<CTA_GROUP, COLLECTOR_USAGE::A_LASTUSE>(acc1_tmem, b_desc, a1_desc, i_desc, sfb_tmem, sfa1_tmem, 1);+ tcgen05_mma_nvfp4<CTA_GROUP>(acc1_tmem, b_desc, a1_desc, i_desc, sfb_tmem, sfa1_tmem, 1);}++ // signal mainloop done+ if (do_commit)+ tcgen05_commit_mcast<CTA_GROUP>(mainloop_mbar_addr + outer_stage1 * 8, cta_mask);}// signal MMA done⋯ 1 unchanged linesinner_stage = (inner_stage + 1) % NUM_STAGES;if (inner_stage == 0)tma_phase ^= 1;- }+ };- // signal mainloop done- tcgen05_commit_mcast<CTA_GROUP>(mainloop_mbar_addr + outer_stage * 8, cta_mask);+ // unroll the 1st iteration to wait for each buffer separately+ issue_mma(0, true, false);- // wait for partial epilogue to finish+ // we use K_dyn to prevent the compiler from unrolling this loop.+ // when using cutlass incantation, adding #pragma unroll 1 to this loop+ // results in segmentation fault.+ for (int iter_k = 1; iter_k < K_dyn / BLOCK_K - 1; iter_k++)+ issue_mma(1, false, false);++ // unroll the last iteration to commit each buffer separately+ issue_mma(1, false, true);++ if constexpr (BLOCK_M == 128) {+ outer_stage0 ^= 1;+ if (outer_stage0 == 0)+ epilogue_phase_128 ^= 1;+ }if constexpr (BLOCK_M == 256) {- mbarrier_wait(epilogue_mbar_addr + (2 + outer_stage) * 8, epilogue_phase ^ 1);- asm volatile("tcgen05.fence::after_thread_sync;");+ outer_stage0 = (outer_stage0 + 2) % 3;+ outer_stage1 = (outer_stage1 + 2) % 3;}-- outer_stage ^= 1;- if (outer_stage == 0)- epilogue_phase ^= 1;}}}else {// epilogue warps- int stage_id = 0;- int mainloop_phase = 0;+ int stage0 = 0;+ int stage1 = 1;+ int mainloop_phase_128 = 0;+ int mainloop_phase_256[3] = {0, 0, 0};const int epilogue_mbar_addr_ = CTA_GROUP == 2 ? (epilogue_mbar_addr & 0xFEFFFFFF) : epilogue_mbar_addr; // report to CTA0const int off_n = bid_n * BLOCK_N;⋯ 2 unchanged linesint raw_bid_m = blockIdx.y;int group_id = 0;int M, bid_m;+ bool do_2nd_mma = false;float tmp[128];⋯ 5 unchanged lineshalf *C_ptr = args.C_ptr_list[group_id];const int stride_cn = cdiv(M, 16) * 16; // multiple of 16- if (warp_id == 0)- mbarrier_wait(mainloop_mbar_addr + stage_id * 8, mainloop_phase);+ if constexpr (BLOCK_M == 256) {+ find_bid(raw_bid_m, group_id, M, bid_m);+ do_2nd_mma = bid_m * BLOCK_M + 128 < M;+ }++ // stage0+ if (warp_id == 0) {+ if constexpr (BLOCK_M == 128)+ mbarrier_wait(mainloop_mbar_addr + stage0 * 8, mainloop_phase_128);+ if constexpr (BLOCK_M == 256) {+ if (stage0 == 0) {+ mbarrier_wait(mainloop_mbar_addr + 0 * 8, mainloop_phase_256[0]);+ mainloop_phase_256[0] ^= 1;+ }+ else if (stage0 == 1) {+ mbarrier_wait(mainloop_mbar_addr + 1 * 8, mainloop_phase_256[1]);+ mainloop_phase_256[1] ^= 1;+ }+ else {+ mbarrier_wait(mainloop_mbar_addr + 2 * 8, mainloop_phase_256[2]);+ mainloop_phase_256[2] ^= 1;+ }+ }+ }bar_sync<bar_epilogue>(4 * WARP_SIZE);asm volatile("tcgen05.fence::after_thread_sync;");- if constexpr (BLOCK_M == 128) {- tcgen05_ld_32x32b<BLOCK_M>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage_id * BLOCK_M);- asm volatile("tcgen05.wait::ld.sync.aligned;");+ tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage0 * 128);+ asm volatile("tcgen05.wait::ld.sync.aligned;");+ mbarrier_arrive(epilogue_mbar_addr_ + stage0 * 8);- for (int m = 0; m < BLOCK_M / 16; m++) {- const int col = off_m + m * 16;- if (col >= M) break;- stg_16<L2_MOD::EVICT_LAST>(C_ptr + (row * stride_cn + col), tmp + m * 16);- }+ for (int m = 0; m < 128 / 16; m++) {+ const int col = off_m + m * 16;+ if (col >= M) break;+ stg_16<L2_MOD::EVICT_LAST>(C_ptr + (row * stride_cn + col), tmp + m * 16);}if constexpr (BLOCK_M == 256) {- // load the overlapping 128 columns first- // for stage_id=0, this correponds to columns [128,256]- // stage_id=1, [ 0,128]- tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, 128);- asm volatile("tcgen05.wait::ld.sync.aligned;");- mbarrier_arrive(epilogue_mbar_addr_ + (2 + stage_id) * 8);+ if (do_2nd_mma) {+ // stage1+ if (warp_id == 0) {+ if (stage1 == 0) {+ mbarrier_wait(mainloop_mbar_addr + 0 * 8, mainloop_phase_256[0]);+ mainloop_phase_256[0] ^= 1;+ }+ else if (stage1 == 1) {+ mbarrier_wait(mainloop_mbar_addr + 1 * 8, mainloop_phase_256[1]);+ mainloop_phase_256[1] ^= 1;+ }+ else {+ mbarrier_wait(mainloop_mbar_addr + 2 * 8, mainloop_phase_256[2]);+ mainloop_phase_256[2] ^= 1;+ }+ }+ bar_sync<bar_epilogue>(4 * WARP_SIZE);+ asm volatile("tcgen05.fence::after_thread_sync;");- for (int m = 0; m < 128 / 16; m++) {- const int col = off_m + (stage_id ^ 1) * 128 + m * 16;- if (col >= M) break;- stg_16<L2_MOD::EVICT_LAST>(C_ptr + (row * stride_cn + col), tmp + m * 16);- }+ tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage1 * 128);+ asm volatile("tcgen05.wait::ld.sync.aligned;");+ mbarrier_arrive(epilogue_mbar_addr_ + stage1 * 8);- // the remaining 128 columns- // for stage_id=0, this correponds to columns [ 0,128]- // stage_id=1, [128,256]- tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage_id * BLOCK_M);- asm volatile("tcgen05.wait::ld.sync.aligned;");-- for (int m = 0; m < 128 / 16; m++) {- const int col = off_m + stage_id * 128 + m * 16;- if (col >= M) break;- stg_16<L2_MOD::EVICT_LAST>(C_ptr + (row * stride_cn + col), tmp + m * 16);+ for (int m = 0; m < 128 / 16; m++) {+ const int col = off_m + 128 + m * 16;+ if (col >= M) break;+ stg_16<L2_MOD::EVICT_LAST>(C_ptr + (row * stride_cn + col), tmp + m * 16);+ }}+ else {+ // arrive immediately+ mbarrier_arrive(epilogue_mbar_addr_ + stage1 * 8);+ }}- mbarrier_arrive(epilogue_mbar_addr_ + stage_id * 8);- stage_id ^= 1;- if (stage_id == 0)- mainloop_phase ^= 1;+ if constexpr (BLOCK_M == 128) {+ stage0 ^= 1;+ if (stage0 == 0)+ mainloop_phase_128 ^= 1;+ }+ if constexpr (BLOCK_M == 256) {+ stage0 = (stage0 + 2) % 3;+ stage1 = (stage1 + 2) % 3;+ }}{⋯ 3 unchanged lineshalf *C_ptr = args.C_ptr_list[group_id];const int stride_cn = cdiv(M, 16) * 16; // multiple of 16- if (warp_id == 0)- mbarrier_wait(mainloop_mbar_addr + stage_id * 8, mainloop_phase);+ if constexpr (BLOCK_M == 256) {+ find_bid(raw_bid_m, group_id, M, bid_m);+ do_2nd_mma = bid_m * BLOCK_M + 128 < M;+ }++ // stage0+ if (warp_id == 0) {+ if constexpr (BLOCK_M == 128)+ mbarrier_wait(mainloop_mbar_addr + stage0 * 8, mainloop_phase_128);+ if constexpr (BLOCK_M == 256) {+ if (stage0 == 0) {+ mbarrier_wait(mainloop_mbar_addr + 0 * 8, mainloop_phase_256[0]);+ mainloop_phase_256[0] ^= 1;+ }+ else if (stage0 == 1) {+ mbarrier_wait(mainloop_mbar_addr + 1 * 8, mainloop_phase_256[1]);+ mainloop_phase_256[1] ^= 1;+ }+ else {+ mbarrier_wait(mainloop_mbar_addr + 2 * 8, mainloop_phase_256[2]);+ mainloop_phase_256[2] ^= 1;+ }+ }+ }bar_sync<bar_epilogue>(4 * WARP_SIZE);asm volatile("tcgen05.fence::after_thread_sync;");- if constexpr (BLOCK_M == 128) {- tcgen05_ld_32x32b<BLOCK_M>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage_id * BLOCK_M);- asm volatile("tcgen05.wait::ld.sync.aligned;");+ tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage0 * 128);+ asm volatile("tcgen05.wait::ld.sync.aligned;");- for (int m = 0; m < BLOCK_M / 16; m++) {- const int col = off_m + m * 16;- if (col >= M) break;- stg_16<L2_MOD::NONE>(C_ptr + (row * stride_cn + col), tmp + m * 16);- }+ for (int m = 0; m < 128 / 16; m++) {+ const int col = off_m + m * 16;+ if (col >= M) break;+ stg_16<L2_MOD::NONE>(C_ptr + (row * stride_cn + col), tmp + m * 16);}if constexpr (BLOCK_M == 256) {- // load the overlapping 128 columns first- // for stage_id=0, this correponds to columns [128,256]- // stage_id=1, [ 0,128]- tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, 128);- asm volatile("tcgen05.wait::ld.sync.aligned;");- mbarrier_arrive(epilogue_mbar_addr_ + (2 + stage_id) * 8);+ if (do_2nd_mma) {+ // stage1+ if (warp_id == 0) {+ if (stage1 == 0)+ mbarrier_wait(mainloop_mbar_addr + 0 * 8, mainloop_phase_256[0]);+ else if (stage1 == 1)+ mbarrier_wait(mainloop_mbar_addr + 1 * 8, mainloop_phase_256[1]);+ else+ mbarrier_wait(mainloop_mbar_addr + 2 * 8, mainloop_phase_256[2]);+ }+ bar_sync<bar_epilogue>(4 * WARP_SIZE);+ asm volatile("tcgen05.fence::after_thread_sync;");- for (int m = 0; m < 128 / 16; m++) {- const int col = off_m + (stage_id ^ 1) * 128 + m * 16;- if (col >= M) break;- stg_16<L2_MOD::NONE>(C_ptr + (row * stride_cn + col), tmp + m * 16);- }+ tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage1 * 128);+ asm volatile("tcgen05.wait::ld.sync.aligned;");- // the remaining 128 columns- // for stage_id=0, this correponds to columns [ 0,128]- // stage_id=1, [128,256]- tcgen05_ld_32x32b<128>(tmp, cta_rank * BLOCK_N + warp_id * 32, stage_id * BLOCK_M);- asm volatile("tcgen05.wait::ld.sync.aligned;");-- for (int m = 0; m < 128 / 16; m++) {- const int col = off_m + stage_id * 128 + m * 16;- if (col >= M) break;- stg_16<L2_MOD::NONE>(C_ptr + (row * stride_cn + col), tmp + m * 16);+ for (int m = 0; m < 128 / 16; m++) {+ const int col = off_m + 128 + m * 16;+ if (col >= M) break;+ stg_16<L2_MOD::NONE>(C_ptr + (row * stride_cn + col), tmp + m * 16);+ }}}}⋯ 160 unchanged linesconstexpr int sm100_size = 227 * 1024;constexpr int dynamic_size = AB_size + SF_size + 2 * 8; // 1 tma_mbar, 1 mma_mbar- constexpr int static_size = 2 * (1 + BLOCK_M / 128) * 8 + 4; // 2 mainloop_mbar, 2*(BLOCK_M/128) epilogue_mbar, tmem_addr+ constexpr int static_size = 3 * 2 * 8 + 4; // 3 mainloop_mbar, 3 epilogue_mbar, tmem_addrconstexpr int NUM_STAGES = (sm100_size - static_size) / dynamic_size;constexpr int smem_size = dynamic_size * NUM_STAGES + static_size;
scrolls · 469 diff lines total
Best evidence level for this revision: reported
JSON